/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.sql.SQLException;
import java.sql.SQLIntegrityConstraintViolationException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.jms.DeliveryMode;
import javax.jms.Queue;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamWriter;

import org.apache.axiom.attachments.Attachments;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMOutputFormat;
import org.apache.axiom.om.impl.dom.ElementImpl;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.commons.io.IOUtils;
import org.bluestemsoftware.open.eoa.aspect.axiom.util.STAXUtils;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.SOAPBindingImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.Constants;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.FileUtils;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.HTTPRequestEntity;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.ResponseProcessor;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.SOAPFaultContext;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.ResponseProcessor.Work;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.jms.JMSRequestor;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.jms.JMSSender;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.ActionMismatchFaultImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.ActionNotSupportedFaultImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.DuplicateMessageIDFaultImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.InvalidEPRFaultImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.InvalidHeaderFaultImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.OnlyAnonymousAddressSupportedFaultImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.OnlyNonAnonymousAddressSupportedFaultImpl;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.WSAHandler;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa.WSAddressingFaultImpl;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.binding.BindingAction;
import org.bluestemsoftware.specification.eoa.component.binding.BindingFault;
import org.bluestemsoftware.specification.eoa.component.binding.BindingFaultReference;
import org.bluestemsoftware.specification.eoa.component.binding.rt.AcceptedResponseError;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.BindingModule;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointOperationReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EngineRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule.WSSpecification;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation.MEP;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.FaultContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingException;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingMessageReference;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPApplicationFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPBindingConfiguration;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPMustUnderstandFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPReceiverFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPSenderFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPVersionMismatchFault;
import org.bluestemsoftware.specification.eoa.ext.feature.http.client.HTTPClientRequest;
import org.bluestemsoftware.specification.eoa.ext.feature.http.client.HTTPClientResponse;
import org.bluestemsoftware.specification.eoa.ext.feature.http.client.HTTPClientRequest.Method;
import org.bluestemsoftware.specification.eoa.ext.feature.http.client.HTTPClientRequest.RequestEntity;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerRequest;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerResponse;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSA;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSAMessageModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSA.WSA10;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.WSABindingModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.fault.WSAddressingFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationData;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationDataModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.mtom.WSMTOM;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.mtom.WSMTOMMessageModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.mtom.WSMTOM.WSMTOM10;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.mtom.soap.WSMTOMBindingModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPHeader;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPTransportModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.rt.TransportFault;
import org.bluestemsoftware.specification.eoa.system.ManagementContext;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

public abstract class SOAPHTTPBindingImpl extends SOAPBindingImpl {

    private static final Log log = SystemContext.getContext().getSystem().getLog(SOAPHTTPBindingImpl.class);

    private static final String TEMP_DIR = System.getProperty("java.io.tmpdir");

    /**
     * Builds content type string according to SOAP version specific rules.
     * 
     * @param outputFormat
     * @param soapAction
     * @return
     */
    protected abstract String getContentType(OMOutputFormat outputFormat, String soapAction);

    /**
     * Extracts soap action from HTTP request according to SOAP version specific rules.
     * @param request
     * @return
     */
    protected abstract String getSOAPAction(HTTPServerRequest request);

    /**
     * Extracts media type from HTTP request according to SOAP version specific rules.
     * 
     * @param contentType
     *        value of Content-Type header
     * @return
     * @throws SOAPFault
     */
    protected abstract String getMediaType(String contentType) throws SOAPFault;

    /**
     * Inspects partner response to determine if it contains a soap protocol fault, i.e. a
     * version mismatch or must understand fault.
     * 
     * @param soapEnvelope
     * @return
     */
    protected abstract boolean isSOAPProtocolFault(SOAPEnvelope soapEnvelope);

    /**
     * Returns soap action and adds an HTTP header according to SOAP version specific rules.
     * 
     * @param httpHeaders
     */
    protected abstract String setSOAPAction(Map<String, String> httpHeaders);

    private Map<String, Map<String, String>> httpHeaders;

    public SOAPHTTPBindingImpl(SOAPBindingConfiguration configuration, Set<String> features) {
        super(configuration, features);
    }

    @Override
    public void spi_init(Set<ManagementContext> managementContexts) throws SOAPBindingException {

        super.spi_init(managementContexts);

        // cache values for required http headers and optionally defined
        // header values

        httpHeaders = new HashMap<String, Map<String, String>>();
        if (!binding.isReusable()) {
            Interface intrface = binding.getInterfaceReference().getReferencedComponent();
            for (InterfaceAction ia : intrface.getInterfaceActions()) {
                BindingAction ba = binding.getBindingAction(ia.getAction());
                if (ba == null) {
                    continue;
                } else {
                    cacheRequiredHTTPHeaders(ia.getAction(), ba);
                }
            }
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPHTTPBindingRT$Provider#spi_receiveAction(org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerRequest,
     *      org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerResponse,
     *      org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction,
     *      org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference)
     */
    public void spi_receiveAction(HTTPServerRequest request, HTTPServerResponse response, Direction direction, EndpointReference epr) {

        log.trace("receiveAction begin");

        // the intent of the logic within this method is to 'accept' the message for
        // processing, see javadoc on SOAPHTTPBindingRT for a detailed description of
        // rules which define the term

        SOAPEnvelope soapEnvelope = null;
        WSMTOMMessageModule mtomMessageModule = null;

        try {

            // if this is a SOAP 1.1 binding and method is not post, as required
            // by the basic profile, return a 405 status code. if version is 1.2,
            // although wsdl 2.0 spec allows GET if soap MEP is 'soap-response',
            // i.e. request is serialized to iri and/or as a query string in
            // http request, we do not support this. this binding requires WSA
            // feature. iri serialization precludes the use of soap header blocks.
            // requesting node should therefore use an endpoint bound to POX HTTP
            // binding instance instead

            if (request.getMethod().equals("POST") == false) {
                log.debug("get method not allowed");
                response.setStatus(Constants.SC_METHOD_NOT_ALLOWED);
                response.trace();
                return;
            }

            // extract media type from content type header. if media type is un-
            // expected, soap version specific descendant throws a soap fault.
            // return 415 status code as required by basic profile

            String mediaType = null;
            try {
                String ctype = request.getContentType();
                ctype = ctype == null ? Constants.DEFAULT_CONTENT_TYPE : ctype;
                mediaType = getMediaType(ctype);
            } catch (SOAPSenderFault se) {
                if (log.isDebugEnabled()) {
                    log.debug("Unsupported media type. " + se.getFaultReason());
                }
                response.setStatus(Constants.SC_UNSUPPORTED_MEDIA_TYPE);
                response.trace();
                return;
            }

            // retrieve input stream from request. check if content is encoded
            // using gzip, which is currently the only encoding we support

            InputStream is = null;
            String ce = request.getHeader(Constants.HEADER_CONTENT_ENCODING);
            if (ce != null) {
                if (ce.equalsIgnoreCase("gzip")) {
                    try {
                        is = new GZIPInputStream(request.getInputStream());
                    } catch (IOException ie) {
                        log.error("Error decoding content. " + ie.getMessage());
                        response.setStatus(Constants.SC_BAD_REQUEST);
                        response.trace();
                        return;
                    }
                } else {
                    if (log.isDebugEnabled()) {
                        log.debug("Content-Encoding '" + ce + "' not supported.");
                    }
                    response.setStatus(Constants.SC_BAD_REQUEST);
                    response.trace();
                    return;
                }
            } else {
                is = request.getInputStream();
            }

            // if media type implies mtom, then we create an attachments object, which
            // leverages the axiom mtom implementation

            Attachments attachments = null;
            if (mediaType.equals(Constants.SOAP_MTOM_MEDIA_TYPE)) {
                String contentType = request.getContentType();
                boolean cacheAttachments = managedConfiguration.getLargeAttachmentSize() > -1;
                String threshold = Integer.toString(managedConfiguration.getLargeAttachmentSize());
                attachments = new Attachments(is, contentType, cacheAttachments, TEMP_DIR, threshold);
                mtomMessageModule = new WSMTOMMessageModule(WSMTOM.WSMTOM10.NAMESPACE_URI);
            }

            // parse request into a soap envelope. if envelope is malformed or if
            // soap version employed fails to match expected version, method throws
            // a soap fault

            soapEnvelope = buildSOAPEnvelope(is, attachments, getCharacterSetEncoding(request));

            // build method returns null if http body is empty, i.e. if stream is
            // zero length

            if (soapEnvelope == null) {
                throw new SOAPSenderFault(getSOAPVersion(), "The request was empty.");
            }

            // attempt to retrieve the payload here, we want to make sure that the
            // start tag of payload element, if not empty, is not malformed

            soapEnvelope.getBody().getFirstElement();

        } catch (Throwable t) {

            // if a soap version mismatch fault was thrown, this requires special
            // handling according to spec

            if (t instanceof SOAPVersionMismatchFault) {
                try {
                    SOAPEnvelope faultEnvelope = buildSOAPFaultEnvelope(epr, (SOAPFault)t);
                    WSAHandler wsaHandler = WSAHandler.getInstance(WSA10.NAMESPACE_URI);
                    String relatesTo = WSA10.UNSPECIFIED_MESSAGE;
                    String action = ((SOAPFault)t).getAction();
                    wsaHandler.handleWSAFaultResponse(null, relatesTo, action, faultEnvelope);
                    OMOutputFormat omFormat = new OMOutputFormat();
                    omFormat.setDoOptimize(false);
                    omFormat.setCharSetEncoding(Constants.UTF_8);
                    omFormat.setAutoCloseWriter(true);
                    omFormat.setSOAP11(getSOAPVersion() == SOAPVersion.SOAP11);
                    response.setContentType(getContentType(omFormat, null));
                    response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                    OutputStream out = response.getOutputStream();
                    faultEnvelope.serializeAndConsume(out, omFormat);
                } catch (Exception ex) {
                    log.error("caught throwable while returning soap fault response. " + ex);
                    response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                }
                response.trace();
                return;
            }

            // if we end up here, this implies that message was NOT 'accepted' for
            // processing. the status code MUST be set to 4xx and the response body
            // must be empty

            String statusReason = null;

            if (!(t instanceof SOAPFault)) {
                statusReason = "Unable to read request";
                log.error("caught throwable while receiving action addressed to epr "
                        + epr.getFragmentIdentifier()
                        + ". "
                        + t);
            } else {
                statusReason = ((SOAPFault)t).getFaultReason();
                if (log.isDebugEnabled()) {
                    log.debug("caught soap fault while receiving action addressed to epr "
                            + epr.getFragmentIdentifier()
                            + ". "
                            + ((SOAPFault)t).toString());
                }
            }

            try {
                response.sendError(Constants.SC_BAD_REQUEST, statusReason);
            } catch (IOException ie) {
                log.error("error sending http error response. " + ie);
            }

            response.trace();
            return;

        }

        // if we make it this far, we have, according to SOAP MEP state model,
        // successfully transitioned from the 'init' state to the 'receiving'
        // state, i.e. the message has been 'accepted for processing.' we now
        // differentiate the logic according to whether message represents an
        // action defined on my service or defined on a partner service

        if (direction == Direction.IN) {
            receiveMyServiceRequest(request, response, epr, mtomMessageModule, soapEnvelope);
        } else {
            receivePartnerServiceResponse(request, response, null, epr, mtomMessageModule, soapEnvelope);
        }

        response.trace();
        
        log.trace("receiveAction end");

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.SOAPHTTPProtocol$Provider#spi_sendAction(org.bluestemsoftware.specification.eoa.component.engine.EndpointReference,
     *      org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext)
     */
    public void spi_sendAction(EndpointReference epr, ActionContext actionContext) throws SystemFault {
        log.trace("sendAction begin");

        ApplicationRT app = epr.getRootComponent().getApplication();
        EndpointActionReference ear = epr.getEndpointActionReference(actionContext.getAction());

        Direction direction = null;
        if (ear == null) {
            if (actionContext.getAction().equals(SystemFault.ACTION)) {
                direction = Direction.OUT;
            } else {
                throw new SystemFault(app, "Action '"
                        + actionContext.getAction()
                        + "' is undefined on referenced endpoint "
                        + epr.getEndpoint().getFragmentIdentifier());
            }
        } else {
            direction = ear.getReferencedComponent().getInterfaceAction().getDirection();
        }

        try {

            if (direction == Direction.OUT) {
                sendMyServiceResponse(epr, actionContext);
            } else {
                sendPartnerServiceRequest(epr, (MessageContext)actionContext);
            }

        } catch (Throwable th) {

            // if we end up here, the implication is that message was not accepted
            // by receiver. we translate into a system fault and throw back to
            // engine and allow it to handle it

            SystemFault systemFault = null;

            if (th instanceof SystemFault) {
                if (log.isDebugEnabled()) {
                    log.debug("caught system fault while sending action "
                            + actionContext.getAction()
                            + " defined on epr "
                            + epr.getFragmentIdentifier()
                            + ". "
                            + ((SystemFault)th).getFaultReason());
                }
                systemFault = (SystemFault)th;
            } else if (th instanceof BindingFaultRT) {
                if (log.isDebugEnabled()) {
                    log.debug("caught binding fault while sending action "
                            + actionContext.getAction()
                            + " defined on epr "
                            + epr.getFragmentIdentifier()
                            + ". "
                            + ((BindingFaultRT)th).getFaultReason());
                }
                QName faultCode = ((BindingFaultRT)th).getFaultName();
                systemFault = new SystemFault(app, faultCode, ((BindingFaultRT)th).getFaultReason());
            } else {
                log.error("caught throwable while sending action "
                        + actionContext.getAction()
                        + " defined on epr "
                        + epr.getFragmentIdentifier()
                        + ". "
                        + th);
                systemFault = new SystemFault(app, "Caught throwable sending action.", th);
            }

            throw systemFault;

        }

        log.trace("sendAction end");
    }

    /*
     * action represents a request addressed to a service hosted by target engine. message
     * direction is in and all supported meps are valid
     */
    private void receiveMyServiceRequest(HTTPServerRequest request, HTTPServerResponse response, EndpointReference epr, WSMTOMMessageModule mtomMessageModule, SOAPEnvelope requestEnvelope) {

        log.trace("receiveMyServiceRequest begin");

        // note that invocation of this method implies that we have 'accepted'
        // the request for processing

        MessageModules requestModules = null;
        WSAMessageModule wsammIn = null;
        EndpointActionReference ear = null;
        EndpointOperationReference eor = null;
        MEP mep = null;
        JMSRequestor jmsRequestor = null;

        try {

            log.debug("receiving my service request");

            // attempt to engage ws-addressing. if this fails, a soap fault
            // is thrown by handler

            WSAHandler wsaHandler = WSAHandler.getInstance(requestEnvelope);
            wsammIn = wsaHandler.handleMyServiceRequest(requestEnvelope);
            String wsaNamespace = wsammIn.getNamespace();
            requestModules = new MessageModules(wsammIn);

            // attempt to register the message. if id is not unique, a sql
            // exception is thrown, i.e. an index constraint violation

            try {
                messageDAO.insertMessage(wsammIn.getMessageID(), true);
            } catch (SQLIntegrityConstraintViolationException se) {
                throw new DuplicateMessageIDFaultImpl(getSOAPVersion(), wsammIn);
            }

            // when receiving an action for my service, the action must be
            // declared

            String action = requestModules.getAction();

            ear = epr.getEndpointActionReference(action);
            if (ear == null) {
                throw new ActionNotSupportedFaultImpl(getSOAPVersion(), wsammIn, action);
            }

            eor = (EndpointOperationReference)ear.getParent();
            mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();

            // this is the initial message in exchange, we need to verify that soap
            // action, if defined, is consistent with the wsa action value. it should
            // be undefined, i.e. it's unused within our framework

            String soapAction = getSOAPAction(request);
            if (soapAction != null && !soapAction.equals(action)) {
                if (log.isDebugEnabled()) {
                    log.debug("SOAPAction '"
                            + soapAction
                            + "' does not match wsa:Action '"
                            + ear.getAction()
                            + "'. Throwing ActionMismatchFault.");
                }
                throw new ActionMismatchFaultImpl(getSOAPVersion(), wsammIn, action, soapAction);
            }

            org.bluestemsoftware.specification.wsa.EndpointReference rt = wsammIn.getReplyTo();
            org.bluestemsoftware.specification.wsa.EndpointReference ft = wsammIn.getFaultTo();
            boolean isBlockingRequest = rt.getAddress().equals(WSA10.ANONYMOUS_URI);
            isBlockingRequest = isBlockingRequest && mep == MEP.IN_OUT;

            // retrieve version of ws-addressing spec employed to encode request
            // headers and check to see that policy was complied with by client.
            // note that api allows for support of future versions of spec

            WSABindingModule wsabm = ear.getBindingModule(WSABindingModule.class);
            WSA wsa = (WSA)wsabm.getSupportedVersions().get(wsaNamespace);
            WSSpecification.Policy wsaPolicy = wsa.getPolicy().iterator().next();

            Boolean isAnonymousResponses = ((WSA10.Policy)wsaPolicy).isAnonymousResponses();
            if (isAnonymousResponses != null) {
                if (isAnonymousResponses) {
                    if (!isBlockingRequest) {
                        throw new OnlyAnonymousAddressSupportedFaultImpl(getSOAPVersion(), wsammIn, WSA.REPLY_TO);
                    }
                } else {
                    if (isBlockingRequest) {
                        throw new OnlyNonAnonymousAddressSupportedFaultImpl(getSOAPVersion(), wsammIn,
                                WSA.REPLY_TO);
                    }
                }
            }

            // if mep is robust in only and faults are to be returned anonymously, client
            // would block for a fault response that may never be returned. we dissallow
            // this configuration, i.e. to avoid the associated performance degradation

            if (mep == MEP.ROBUST_IN_ONLY) {
                if (ft != null) {
                    if (ft.getAddress().equals(WSA10.ANONYMOUS_URI)) {
                        log.debug("robust-in-only MEP and FaultTo is anonymous");
                        throw new InvalidHeaderFaultImpl(getSOAPVersion(), wsammIn, "FaultTo");
                    }
                } else {
                    if (rt.getAddress().equals(WSA10.ANONYMOUS_URI)) {
                        log.debug("robust-in-only MEP and FaultTo is null and ReplyTo is anonymous");
                        throw new InvalidHeaderFaultImpl(getSOAPVersion(), wsammIn, "ReplyTo");
                    }
                }
            }

            // if client is blocking for a normal response and fault is returned async,
            // client would block until timed out. or if client is blocking for a fault
            // response and normal response is returned async, client would block until
            // timed out. in either case we dissallow this, i.e. to avoid the associated
            // performance degradation

            if (mep == MEP.IN_OUT) {
                if (ft != null) {
                    if (isBlockingRequest) {
                        if (!ft.getAddress().equals(WSA10.ANONYMOUS_URI)) {
                            log.debug("ReplyTo anonymous, but FaultTo non-anonymous");
                            throw new InvalidHeaderFaultImpl(getSOAPVersion(), wsammIn, "FaultTo");
                        }
                    } else {
                        if (ft.getAddress().equals(WSA10.ANONYMOUS_URI)) {
                            log.debug("FaultTo anonymous, but ReplyTo non-anonymous");
                            throw new InvalidHeaderFaultImpl(getSOAPVersion(), wsammIn, "FaultTo");
                        }
                    }
                }
            }

            // the wsa:To value should match address used to send request, i.e. if
            // user passes us a bogus to epr, super won't be able to use the wsa
            // header to match message deserialized from queue with an epr

            String httpTo = request.getRequestURL().toString();
            if (!httpTo.equals(wsammIn.getTo())) {
                log.debug("request URL '" + httpTo + "' doesn't match wsa:To '" + wsammIn.getTo() + "'");
                throw new InvalidEPRFaultImpl(getSOAPVersion(), wsammIn, "To");
            }

            // ws-addressing is engaged, i.e. no wsa faults should be thrown past
            // this point. now that message modules collection has been seeded with
            // required wsa message module, add optional mtom module

            if (mtomMessageModule != null) {
                requestModules.putModule(mtomMessageModule);
            }

            // verify that if binding contains abstract http header decls
            // marked as required, that they exist and if user has defined
            // a value within provider configuration, verifiy it matches

            checkRequiredHTTPHeaders(action, null, request);

            // iterate over declared soap headers and verify that, if required,
            // the header is defined on message. ws-a headers have already been
            // extracted and are therefore assumed to exist

            checkRequiredSOAPHeaders(getConcreteMessageDefinition(wsammIn.getAction()), requestEnvelope);

            // execute binding modules. note that modules natively implemented by
            // this binding are a no-op

            MessageModule mm = null;
            for (BindingModule bm : ear.getBindingModules()) {
                mm = bm.handleMyServiceRequest(requestModules, (Content)requestEnvelope);
                if (mm != null) {
                    requestModules.putModule(mm);
                }
            }

            // process remaining headers with application data handler. it
            // will extract headers targeted to ultimate destination including
            // any with role matching role implemented by referenced service.
            // if a header is flagged as mustUnderstand, the handler will
            // exclude it from module unless it was defined via wsad policy

            Map<QName, Set<ApplicationData>> ad = readApplicationData(ear, requestEnvelope);
            if (ad != null) {
                requestModules.getModule(ApplicationDataModule.class).setApplicationData(ad);
            }

            // iterate over remaining headers and throw a must understand
            // fault if any headers remain that we don't understand

            checkMustUnderstand(requestEnvelope);

            // serialize message content to a file to avoid buffering entire
            // message to memory which is limitation of jms api. we use
            // message id as file name, which will be used by jms receiver
            // to retrieve from file system

            String fileName = FileUtils.createFileName(wsammIn.getMessageID());
            File requestContent = new File(dataDirectory, fileName);

            // if payload is null, i.e. it's an empty request. we create an
            // empty file, i.e. which will allow user to manage the request
            // files within the binding var dir, i.e. browsing/deleting,
            // etc ... if jms receiver can't find file, it quietly logs a
            // debug message

            OMElement payload = requestEnvelope.getBody().getFirstElement();

            if (payload != null) {
                OutputStream out = null;
                XMLStreamWriter writer = null;
                try {
                    out = new FileOutputStream(requestContent);
                    writer = STAXUtils.createXMLStreamWriter(out, Constants.UTF_8);
                    payload.buildNext(); // kids won't serialize otherwise
                    payload.serializeAndConsume(writer);
                } finally {
                    if (out != null) {
                        out.flush();
                        IOUtils.closeQuietly(out);
                    }
                    if (writer != null) {
                        writer.close();
                    }
                }
            } else {
                requestContent.createNewFile();
            }

            // if this is a blocking request, we overwrite anonymous replyTo
            // and faultTo with name of temporary queue used by requestor
            // before persisting the module to dbms. handleMyServiceResponse
            // detects by checking url for jms scheme

            if (isBlockingRequest) {
                jmsRequestor = (JMSRequestor)jmsRequestors.borrowObject();
                String replyTo = "jms:" + jmsRequestor.getTemporaryQueue().getQueueName();
                Set<Element> rp = rt.getReferenceParameters();
                org.bluestemsoftware.specification.wsa.EndpointReference temp;
                temp = new org.bluestemsoftware.specification.wsa.EndpointReference(replyTo, rp);
                wsammIn.setReplyTo(temp);
                if (ft != null) {
                    rp = ft.getReferenceParameters();
                    temp = new org.bluestemsoftware.specification.wsa.EndpointReference(replyTo, rp);
                    wsammIn.setFaultTo(temp);
                }
            }

            // persist message modules to dbms before passing execution off
            // to a separate thread

            for (MessageModule messageModule : requestModules.values()) {
                messageModuleDAO.insertMessageModule(wsammIn.getMessageID(), true, messageModule);
            }

            // write message to queue where it will be picked up by jms
            // receiver on separate thread and processed by read message
            // method on super

            log.debug("writing message to queue");

            JMSSender jmsSender = null;
            try {
                jmsSender = (JMSSender)jmsSenders.borrowObject();
                javax.jms.Message msg = jmsSender.createMessage();
                msg.setJMSCorrelationID(wsammIn.getMessageID());
                msg.setStringProperty(Constants.MESSAGE_ID_PROPERTY, wsammIn.getMessageID());
                msg.setBooleanProperty(Constants.MY_SERVICE_PROPERTY, true);
                jmsSender.send(msg, receiveQueue, 0, DeliveryMode.PERSISTENT);
            } finally {
                jmsSenders.returnObject(jmsSender);
            }

            // if anonymous, block for response using jmsRequestor. the response
            // is a fully processed soap envelope (unless mep is robust-in-only
            // and no fault occurred, in which case response is empty. note that
            // content type is passed as a message property which indicates char
            // encoding used to serialize chars to bytes

            if (isBlockingRequest) {

                long timeout = 0;
                if (managedConfiguration.getRequestorTimeout(eor.getEndpointOperationName()) == null) {
                    timeout = managedConfiguration.getDefaultRequestorTimeout();
                } else {
                    timeout = managedConfiguration.getRequestorTimeout(eor.getEndpointOperationName());
                }

                if (log.isDebugEnabled()) {
                    log.debug("response is anonymous. requestor blocking "
                            + timeout
                            + " milliseconds for engine response.");
                }

                String jmsCorrellationID = wsammIn.getMessageID();
                javax.jms.Message msg = jmsRequestor.receive(jmsCorrellationID, timeout);

                // set replyTo and faultTo back to anonymous uri in case a fault
                // is thrown from this point forward, i.e. so sendMyServiceResponse
                // will not attempt to return response via jms

                wsammIn.setReplyTo(rt);
                wsammIn.setFaultTo(ft);

                // if msg is null, we timed out waiting for response. throw a fault
                // so that client will receive notification

                if (msg == null) {
                    throw new SOAPReceiverFault(getSOAPVersion(), "Binding timed-out waiting for response.");
                }

                log.debug("requestor received response");

                boolean gzipResponse = msg.getBooleanProperty(Constants.GZIP_RESPONSE_PROPERTY);
                boolean isFaultResponse = msg.getBooleanProperty(Constants.IS_FAULT_RESPONSE_PROPERTY);
                String responseContentType = msg.getStringProperty(Constants.CONTENT_TYPE_PROPERTY);
                String responseAction = msg.getStringProperty(Constants.ACTION_PROPERTY);

                // set user defined response http headers keying off of the action
                // value which was passed to us via a jms property

                Map<String, String> responseHeaders = httpHeaders.get(responseAction);

                if (responseHeaders != null) {
                    for (Map.Entry<String, String> entry : responseHeaders.entrySet()) {
                        String name = entry.getKey();
                        String value = entry.getValue();
                        if (response.containsHeader(name) == false) {
                            response.addHeader(name, value);
                        } else {
                            log.debug("not overwriting existing http header '"
                                    + name
                                    + "' with user defined value '"
                                    + value
                                    + "'.");
                        }
                    }
                }

                // set the response status. note that if response contains a soap
                // fault, basic profile requires that we set status to 500

                response.setContentType(responseContentType);
                if (isFaultResponse) {
                    log.debug("engine returned fault response");
                    response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                } else {
                    log.debug("engine returned normal response");
                    response.setStatus(Constants.SC_OK);
                }

                String responseFileName = msg.getStringProperty(Constants.FILE_NAME_PROPERTY);
                File responseMessageContent = new File(responseFileName);

                OutputStream out = response.getOutputStream();
                if (gzipResponse) {
                    response.addHeader(Constants.HEADER_CONTENT_ENCODING, Constants.COMPRESSION_GZIP);
                    out = new GZIPOutputStream(response.getOutputStream());
                }

                InputStream in = null;
                try {
                    in = new FileInputStream(responseMessageContent);
                    IOUtils.copy(in, out);
                } finally {
                    IOUtils.closeQuietly(in);
                    if (!responseMessageContent.delete()) {
                        log.error("error deleting processed response "
                                + responseMessageContent.getAbsolutePath()
                                + " from file system");
                    }
                }

            } else {

                log.debug("reply will be returned async. setting response status to 202");

                // mep is in-only or mep is in-out and response is returned async,
                // or mep is robust-in-only and fault response, if any, is returned
                // async. in any case, set status to 202 to indicate request was
                // accepted for processing

                response.setStatus(Constants.SC_ACCEPTED);

            }

        } catch (Throwable th) {

            if (th instanceof WSAddressingFault) {
                wsammIn = ((WSAddressingFaultImpl)th).getOffendingMessageModule();
            }

            SOAPFault sf = null;

            if (th instanceof SOAPFault) {
                sf = (SOAPFault)th;
                if (log.isDebugEnabled()) {
                    log.debug("caught soap fault while receiving request action "
                            + wsammIn.getAction()
                            + " defined on endpoint referenced by component "
                            + epr.getFragmentIdentifier()
                            + ". "
                            + ((SOAPFault)th).getFaultReason());
                }
            } else {
                if (managedConfiguration.isEmbeddStackTraceInFaults()) {
                    sf = new SOAPReceiverFault(getSOAPVersion(), th);
                } else {
                    sf = new SOAPReceiverFault(getSOAPVersion(), th.getMessage());
                }
                log.error("caught throwable while receiving request action "
                        + wsammIn.getAction()
                        + " defined on endpoint referenced by component "
                        + epr.getFragmentIdentifier()
                        + ". "
                        + th);
            }

            ear = epr.getEndpointActionReference(wsammIn.getAction());

            if (ear != null) {
                eor = (EndpointOperationReference)ear.getParent();
                mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();
            }

            // basic profile requires an empty body in response to one way mep,
            // with an accepted code, even if we failed to process the request

            if (mep != null && mep == MEP.IN_ONLY) {
                log.debug("MEP is in-only. No fault response allowed.");
                response.setStatus(Constants.SC_ACCEPTED);
                return;
            }

            // determine how fault response will be returned. note that
            // must understand faults are always returned anonymously,
            // regardless of how response epr(s) is/are manifested

            boolean isAnonymous = false;

            if (wsammIn.getFaultTo() != null) {

                if (wsammIn.getFaultTo().getAddress().equals(WSA10.ANONYMOUS_URI)) {
                    isAnonymous = true;
                } else {
                    if (th instanceof SOAPMustUnderstandFault) {
                        wsammIn.setFaultTo(WSA10.DEFAULT_REPLY_TO);
                        isAnonymous = true;
                    }
                }

            } else {

                // if replyTo is null, then the epr was malformed, otherwise
                // it would have defaulted to anon. set to anonymous uri, so
                // that response can be processed

                if (wsammIn.getReplyTo() == null) {
                    wsammIn.setReplyTo(WSA10.DEFAULT_REPLY_TO);
                    isAnonymous = true;
                } else {
                    if (wsammIn.getReplyTo().getAddress().equals(WSA10.ANONYMOUS_URI)) {
                        isAnonymous = true;
                    } else {
                        if (th instanceof SOAPMustUnderstandFault) {
                            wsammIn.setReplyTo(WSA10.DEFAULT_REPLY_TO);
                            isAnonymous = true;
                        }
                    }
                }

            }

            // if fault is to be returned asynchronously, process the response on a
            // separate thread, i.e. so that we can accept request before returning
            // response. note that basic profile (R1151) prohibits us from returning
            // wsa fault response over the back channel if response is non-anonymous

            if (!isAnonymous) {
                if (wsammIn.getMessageID().equals(WSA10.UNSPECIFIED_MESSAGE)) {
                    log.debug("unable to return wsa fault response async. no message id.");
                    response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                    return;
                }
                ResponseProcessor responseProcessor = null;
                try {
                    responseProcessor = (ResponseProcessor)responseProcessorPool.borrowObject();
                    responseProcessor.setWork(new WorkImpl(sf, epr, ear, requestModules));
                    response.setStatus(Constants.SC_ACCEPTED);
                } catch (Exception ex) {
                    log.error("caught throwable returning async binding fault. " + ex);
                    response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                } finally {
                    try {
                        responseProcessorPool.returnObject(responseProcessor);
                    } catch (Exception fatchance) {
                    }
                }
                return;
            }

            // if ws-addressing was not engaged, this is considered fatal,
            // do not process response logic. note that basic profile
            // (R1151) prohibits us from returning fault response over the
            // back channel if reply is non-anonymous

            if (th instanceof WSAddressingFault) {

                SOAPEnvelope faultEnvelope = buildSOAPFaultEnvelope(epr, (SOAPFault)th);
                OMOutputFormat omFormat = new OMOutputFormat();
                omFormat.setDoOptimize(false);
                omFormat.setCharSetEncoding(Constants.UTF_8);
                omFormat.setAutoCloseWriter(true);
                omFormat.setSOAP11(getSOAPVersion() == SOAPVersion.SOAP11);
                String contentType = getContentType(omFormat, null);

                WSAHandler wsaHandler = WSAHandler.getInstance(WSA10.NAMESPACE_URI);
                String mid = wsammIn.getMessageID();
                try {
                    wsaHandler.handleWSAFaultResponse(null, mid, sf.getAction(), faultEnvelope);
                    response.setContentType(contentType);
                    response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                    OutputStream out = response.getOutputStream();
                    faultEnvelope.serializeAndConsume(out, omFormat);
                } catch (Exception ex) {
                    log.error("caught throwable returning wsa fault response. " + ex);
                    response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                }
                return;
            }

            // ws-addressing was engaged. run the binding fault response through
            // handle response logic. http response is passed by reference, i.e.
            // binding fault response is processed on same thread as request

            try {
                WSAMessageModule wsammOut = new WSAMessageModule(sf.getAction());
                wsammOut.setRelatesTo(wsammIn.getMessageID());
                MessageModules responseModules = new MessageModules(wsammOut);
                SOAPFaultContext sfc = new SOAPFaultContext(response, requestModules, responseModules, sf);
                sendMyServiceResponse(epr, sfc);
            } catch (Throwable thr) {
                log.error("caught throwable while processing soap fault. " + thr);
                response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
            }

        } finally {
            if (jmsRequestor != null) {
                try {
                    jmsRequestors.returnObject(jmsRequestor);
                } catch (Exception fatchance) {
                }
            }
        }

        log.trace("receiveMyServiceRequest end");

    }

    /*
     * action represents a response from a partner service referenced by target engine. message
     * direction is out and mep is in-out or robust in-only
     */
    private void receivePartnerServiceResponse(HTTPServerRequest request, HTTPServerResponse serverResponse, HTTPClientResponse clientResponse, EndpointReference epr, WSMTOMMessageModule mtommm, SOAPEnvelope responseEnvelope) {

        log.trace("receivePartnerServiceResponse begin");

        // invocation of this method implies message has been accepted by receiver,
        // i.e. if call originates via spi_receiveAction method we are receiver. if
        // call originates via sendPartnerServiceRequest method, i.e. it's an anon
        // response, partner is receiver and has implicitly accepted the request

        ApplicationRT app = ((EngineRT)epr.getRootComponent()).getApplication();

        MessageModules responseModules = null;
        MessageModules requestModules = null;
        WSAMessageModule wsamm = null;
        List<Content> content = null;

        try {

            log.debug("receiving partner service response");

            // attempt to engage ws-addressing. if this fails, a soap fault
            // is thrown by handler. if not engaged, we cannot continue

            String action = null;
            WSAHandler wsaHandler = WSAHandler.getInstance(responseEnvelope);
            try {
                wsamm = wsaHandler.handlePartnerServiceResponse(responseEnvelope);
                responseModules = new MessageModules(wsamm);
                action = wsamm.getAction();
            } catch (WSAddressingFault wsaf) {
                if (log.isDebugEnabled()) {
                    log.debug("unable to process partner response from epr "
                            + epr.getFragmentIdentifier()
                            + ". failed to engage ws-addressing. "
                            + wsaf.getFaultReason());
                }
                if (serverResponse != null) {
                    // spec requires that the status must be set to 2xx
                    serverResponse.setStatus(Constants.SC_ACCEPTED);
                }
                return;
            }

            // if this is an anonymous partner response passed to us by
            // sendPartnerServiceRequest method, we overwrite the
            // anonymous To epr defined on wsa module with the epr
            // fragment identifier. super will use it to map message
            // pulled off of queue with an epr, i.e. so that it can be
            // bound within the readMessage method

            if (wsamm.getTo().equals(WSA10.ANONYMOUS_URI)) {

                wsamm.setTo(epr.getFragmentIdentifier().toString());

            } else {

                // response was returned asychronously. the to epr should
                // match url on http request, i.e. if user passes us a
                // bogus wsa:To address, super won't be able to use the
                // wsa to header to match message deserialized from queue
                // with an epr object

                String httpTo = request.getRequestURL().toString();
                if (!httpTo.equals(wsamm.getTo())) {
                    if (log.isDebugEnabled()) {
                        log.debug("unable to process partner response from epr "
                                + epr.getFragmentIdentifier()
                                + ". Failed to engaged ws-addressing. "
                                + "request URL '"
                                + httpTo
                                + "' doesn't match wsa:To '"
                                + wsamm.getTo()
                                + "'");
                    }
                    // spec requires that the status must be set to 2xx
                    serverResponse.setStatus(Constants.SC_ACCEPTED);
                    return;
                }

            }

            // using relatesTo message id defined on wsa header, look up and
            // attempt to correllate with an outstanding request. if we fail
            // to correllate, we cannot continue

            String requestMessageID = wsamm.getRelatesTo();
            requestModules = messageModuleDAO.selectMessageModules(requestMessageID, false);

            if (requestModules == null) {
                log.warn("unable to process partner response from epr "
                        + epr.getFragmentIdentifier()
                        + ". failed to correllate relatesTo '"
                        + requestMessageID
                        + "' with a request.");
                if (serverResponse != null) {
                    // spec requires that the status must be set to 2xx
                    serverResponse.setStatus(Constants.SC_ACCEPTED);
                }
                log.trace("receivePartnerServiceResponse end");
                return;
            }

            // attempt to register the message. if id is not unique, a sql
            // exception is thrown, i.e. an index constraint violation.
            // note that if partner failed to supply a messageid, one was
            // generated by handler

            try {

                messageDAO.insertMessage(wsamm.getMessageID(), false);

            } catch (SQLIntegrityConstraintViolationException se) {

                // fault handling logic depends upon the existence of
                // a message id/record. so, log an error and create a
                // new module with a unique id. a partner supplied msg
                // id on a response is more or less irrelevant and
                // non-standard and therefore shouldn't be a problem

                WSAMessageModule temp = new WSAMessageModule(action);
                temp.setTo(wsamm.getTo());
                temp.setNamespace(wsamm.getNamespace());
                temp.setRelatesTo(wsamm.getMessageID());

                log.debug("partner supplied message id '"
                        + wsamm.getMessageID()
                        + "' defined on response action "
                        + wsamm.getAction()
                        + " is not unique. "
                        + " overwriting with unique id '"
                        + temp.getMessageID()
                        + "'.");

                wsamm = temp;

                messageDAO.insertMessage(wsamm.getMessageID(), false);

            }

            // now that message modules collection has been seeded
            // with required wsa module, add optional mtom module

            if (mtommm != null) {
                responseModules.putModule(mtommm);
            }

            // retrieve message payload which may be null. note that if
            // it's a fault response, payload is extracted from detail
            // element of fault

            QName faultCode = null;
            if (responseEnvelope.getBody().hasFault()) {
                faultCode = SOAPFault.getFaultName((Element)responseEnvelope);
                content = SOAPFault.getFaultDetail((Element)responseEnvelope);
            } else {
                Content temp = (Content)responseEnvelope.getBody().getFirstElement();
                if (temp != null) {
                    content = Arrays.asList(temp);
                }
            }

            // axis (as of version 1.3) does not use predefined wsa soap fault
            // action uri if it's a must understand fault, i.e. it uses an
            // arbitrary declared fault action

            // TODO: file axis bug report and remove code below when fixed

            if (faultCode != null) {
                SOAPVersion sv = getSOAPVersion();
                if (sv == SOAPVersion.SOAP11) {
                    if (faultCode.equals(SOAPMustUnderstandFault.SOAP_11_FAULT_CODE)) {
                        action = SOAPFault.DEFAULT_ACTION;
                        wsamm.setAction(action);
                    }
                } else {
                    if (faultCode.equals(SOAPMustUnderstandFault.SOAP_12_FAULT_CODE)) {
                        action = SOAPFault.DEFAULT_ACTION;
                        wsamm.setAction(action);
                    }
                }
            }

            // verify that if binding contains abstract http header decls
            // marked as required, that they exist and if user has defined
            // a value within provider configuration, verifiy it matches

            checkRequiredHTTPHeaders(action, clientResponse, request);

            // iterate over declared soap headers and verify that, if required,
            // the header is defined on message. ws-a headers have already been
            // extracted and are therefore assumed to exist

            checkRequiredSOAPHeaders(getConcreteMessageDefinition(action), responseEnvelope);

            EndpointActionReference ear = epr.getEndpointActionReference(action);
            List<BindingModule> bindingModules = null;
            if (ear == null) {
                bindingModules = epr.getDefaultBindingModules();
            } else {
                bindingModules = ear.getBindingModules();
            }

            // iterate over binding modules and invoke each. note that modules natively
            // implemented by this binding are a no-op

            for (BindingModule bindingModule : bindingModules) {
                Content msg = (Content)responseEnvelope;
                MessageModule rm = null;
                try {
                    rm = bindingModule.handlePartnerServiceResponse(requestModules, responseModules, msg);
                } catch (Throwable th) {
                    String faultReason = null;
                    if (th instanceof SOAPFault) {
                        faultCode = ((SOAPFault)th).getFaultName();
                        faultReason = ((SOAPFault)th).getFaultReason();
                        if (log.isDebugEnabled()) {
                            log.debug("binding module threw a binding fault. " + faultReason);
                        }
                    } else {
                        faultCode = BindingFaultRT.RECEIVER_ERROR;
                        faultReason = th.getMessage();
                        log.error("binding module threw an unexpected exception. " + th);
                    }
                    throw new AcceptedResponseError(faultCode, faultReason, content);
                }
                if (rm != null) {
                    responseModules.putModule(rm);
                }
            }

            // process remaining headers with application data handler. it
            // will extract headers targeted to ultimate destination including
            // any with role matching role implemented by referenced service.
            // if a header is flagged as mustUnderstand, the handler will
            // exclude it from module unless it was defined via wsad policy

            if (ear != null) {
                Map<QName, Set<ApplicationData>> ad = readApplicationData(ear, responseEnvelope);
                if (ad != null) {
                    responseModules.getModule(ApplicationDataModule.class).setApplicationData(ad);
                }
            }

            // iterate over remaining headers and throw a must understand
            // fault if any found that we don't understand

            checkMustUnderstand(responseEnvelope);

            // now we must handle undeclared actions which is either an
            // undeclared normal response or an undeclared fault action

            if (ear == null) {
                String faultReason = null;
                if (faultCode != null) {
                    faultReason = SOAPFault.getFaultReason((Element)responseEnvelope);
                } else {
                    faultCode = BindingFaultRT.RECEIVER_ERROR;
                    faultReason = "Response action "
                            + wsamm.getAction()
                            + " defined on partner response is undeclared.";
                }
                throw new AcceptedResponseError(faultCode, faultReason, content);
            }

            // so if we make it to this point, we have a declared action which
            // maps to a payload. note that payload could be a normal message,
            // a fault message, or empty. payload will be verified by our soap
            // reader when pulled of queue by super

        } catch (Throwable th) {

            if (th instanceof SOAPMustUnderstandFault) {
                if (serverResponse != null) {

                    // this implies that partner response was sent to us
                    // async, i.e. via spi_receiveAction method, we must
                    // return fault over back channel as required by spec

                    try {
                        SOAPEnvelope faultEnvelope = buildSOAPFaultEnvelope(epr, (SOAPFault)th);
                        WSAHandler wsaHandler = WSAHandler.getInstance(WSA10.NAMESPACE_URI);
                        String relatesTo = wsamm.getMessageID();
                        String action = ((SOAPFault)th).getAction();
                        wsaHandler.handleWSAFaultResponse(null, relatesTo, action, faultEnvelope);
                        OMOutputFormat omFormat = new OMOutputFormat();
                        omFormat.setDoOptimize(false);
                        omFormat.setCharSetEncoding(Constants.UTF_8);
                        omFormat.setAutoCloseWriter(true);
                        omFormat.setSOAP11(getSOAPVersion() == SOAPVersion.SOAP11);
                        serverResponse.setContentType(getContentType(omFormat, null));
                        serverResponse.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                        OutputStream out = serverResponse.getOutputStream();
                        faultEnvelope.serializeAndConsume(out, omFormat);
                    } catch (Throwable t) {
                        log.error("caught throwable while returning soap fault response. " + t);
                        serverResponse.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);
                    }

                    return;

                } else {

                    // we're processing a synchronous partner response, in
                    // which case, no indication is communicated to our
                    // partner, but is instead thrown to engine, i.e.
                    // we just fall through to logic below

                }
            }

            // because response has been accepted for processing, any exceptions
            // caught here must be communicated to engine via receive action method.
            // convert the exception into an accepted response error which will
            // be used to generate an instance of system fault

            AcceptedResponseError are = null;

            if (th instanceof AcceptedResponseError) {
                are = (AcceptedResponseError)th;
            } else {
                QName faultCode = null;
                String faultReason = null;
                if (th instanceof SOAPFault) {
                    faultCode = ((SOAPFault)th).getFaultName();
                    faultReason = ((SOAPFault)th).getFaultReason();
                    if (log.isDebugEnabled()) {
                        log.debug("caught soap fault while processing partner response. " + faultReason);
                    }
                } else {
                    faultCode = BindingFaultRT.RECEIVER_ERROR;
                    faultReason = "Caught throwable while processing partner response. " + th.getMessage();
                    log.error("caught throwable while processing partner response. " + th);
                }
                are = new AcceptedResponseError(faultCode, faultReason, content);
            }

            // if partner is an eoa compliant engine, they may have thrown a
            // system fault, in which case, use existing payload, which should
            // be a system fault message

            boolean useExistingPayload = false;

            if (wsamm.getAction().equals(SystemFault.ACTION)) {
                useExistingPayload = true;
            }

            // translate soap fault into a system fault so that we can use it
            // to generate an instanceof system fault message which contains
            // partner response within fault detail

            if (useExistingPayload == false) {
                SystemFault systemFault = SystemFault.valueOf(app, are);
                Content temp = systemFault.toMessage().getContent();
                Document owner = ((Element)responseEnvelope).getOwnerDocument();
                content = Arrays.asList((Content)owner.importNode(temp, true));
            }

            // payload is a runtime instance of system fault message.
            // change action to system fault action, i.e. engine
            // should never see an undeclared action uri

            wsamm.setAction(SystemFault.ACTION);

        }

        // persist message modules to dbms before passing execution off
        // to a separate thread

        for (MessageModule messageModule : responseModules.values()) {
            try {
                messageModuleDAO.insertMessageModule(wsamm.getMessageID(), false, messageModule);
            } catch (Exception ex) {
                log.error("error persisting message modules. " + ex);
                return;
            }
        }

        // write message to queue where it will be picked up by jms
        // receiver on separate thread and processed by read message
        // method on super

        JMSSender jmsSender = null;
        try {

            log.debug("writing message to queue");

            // serialize message content to a file to avoid buffering entire
            // message to memory as required by jms api. we use message id as
            // file name, which will be used by jms receiver to retrieve from
            // file system

            String fileName = FileUtils.createFileName(wsamm.getMessageID());
            File responseContent = new File(dataDirectory, fileName);

            // if payload is null, i.e. it's an empty response. we create an
            // empty file, i.e. which will allow user to manage the request
            // files within the binding var dir, i.e. browsing/deleting,
            // etc ... if jms receiver can't find file, it quietly logs a
            // debug message

            if (content != null) {
                OutputStream out = null;
                XMLStreamWriter writer = null;
                try {
                    out = new FileOutputStream(responseContent);
                    writer = STAXUtils.createXMLStreamWriter(out, Constants.UTF_8);
                    OMElement temp = (OMElement)content.iterator().next();
                    temp.buildNext(); // kids won't serialize otherwise
                    temp.serializeAndConsume(writer);
                } finally {
                    if (out != null) {
                        out.flush();
                        IOUtils.closeQuietly(out);
                    }
                    if (writer != null) {
                        writer.close();
                    }
                }
            } else {
                responseContent.createNewFile();
            }

            jmsSender = (JMSSender)jmsSenders.borrowObject();
            javax.jms.Message msg = jmsSender.createMessage();
            msg.setStringProperty(Constants.MESSAGE_ID_PROPERTY, wsamm.getMessageID());
            msg.setBooleanProperty(Constants.MY_SERVICE_PROPERTY, false);
            jmsSender.send(msg, receiveQueue, 0, DeliveryMode.PERSISTENT);

        } catch (Throwable th) {
            log.error("caught throwable while receiving partner response. " + th);
        } finally {
            try {
                jmsSenders.returnObject(jmsSender);
            } catch (Exception fatchance) {
            }
        }

        // specificiation requires that we set response status to accepted
        // regardless of whether or not a fault was encountered

        if (serverResponse != null) {
            log.debug("setting response status to 202");
            serverResponse.setStatus(Constants.SC_ACCEPTED);
        }

        log.trace("receivePartnerServiceResponse end");

    }

    /*
     * action represents a response being returned by a locally hosted service. message
     * direction is out and mep is in-out or robust in-only
     */
    protected void sendMyServiceResponse(EndpointReference epr, ActionContext actionContext) throws Exception {

        log.trace("sendMyServiceResponse begin");

        // if action context is an instanceof SOAPFaultContext, then request
        // failed to enter application layer, in which case request modules
        // are passed by reference. otherwise look-up using relates to value

        log.debug("sending my service response");

        ApplicationRT app = ((EngineRT)epr.getRootComponent()).getApplication();
        WSAMessageModule wsammOut = actionContext.getMessageModule(WSAMessageModule.class);
        String requestMessageID = wsammOut.getRelatesTo();
        MessageModules requestModules = null;

        WSAMessageModule wsammIn = null;
        if (actionContext instanceof SOAPFaultContext) {
            requestModules = ((SOAPFaultContext)actionContext).getRequestModules();
        } else {
            requestModules = messageModuleDAO.selectMessageModules(requestMessageID, true);
            if (requestModules == null) {
                throw new SystemFault(app, "Failed to correllate request.");
            }
        }

        wsammIn = requestModules.getModule(WSAMessageModule.class);

        // attempt to persist the message. if id is not unique, a sql
        // exception is thrown

        try {
            messageDAO.insertMessage(wsammOut.getMessageID(), true);
        } catch (SQLIntegrityConstraintViolationException se) {
            throw new SystemFault(app, "Response messageID not unique.");
        }

        HyperTextTransferProtocol http = ((HyperTextTransferProtocol)binding.getUnderlyingProtocol());

        SOAPEnvelope responseEnvelope = null;
        String responseAction = actionContext.getAction();
        List<BindingModule> bindingModules;
        boolean gzipResponse = false;
        boolean isMTOMEnabled = false;
        boolean isFaultResponse = false;

        EndpointActionReference responseEAR = epr.getEndpointActionReference(responseAction);

        if (responseEAR != null) {

            EndpointOperationReference eor = (EndpointOperationReference)responseEAR.getParent();
            MEP mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();

            if (mep == MEP.IN_ONLY) {
                throw new SystemFault(app, "MEP is 'in-only'. No response is allowed.");
            }

            InterfaceAction ia = responseEAR.getReferencedComponent().getInterfaceAction();
            if (ia instanceof InterfaceFaultReference) {
                isFaultResponse = true;
                QName faultName = ((InterfaceFaultReference)ia).getFault().getName();
                String encodings = http.getFaultHTTPContentEncoding(faultName);
                if (encodings != null) {
                    gzipResponse = encodings.contains("gzip");
                }
                SOAPApplicationFault sf = new SOAPApplicationFault(getSOAPVersion(), (FaultContext)actionContext);
                responseEnvelope = buildSOAPFaultEnvelope(epr, sf);
            } else {
                if (mep == MEP.ROBUST_IN_ONLY) {
                    throw new SystemFault(app, "MEP is 'robust-in-only'. Normal response not allowed.");
                }
                QName operationName = eor.getEndpointOperationName();
                String encodings = http.getMessageReferenceHTTPContentEncoding(operationName, Direction.OUT);
                if (encodings != null) {
                    gzipResponse = encodings.contains("gzip");
                }
                responseEnvelope = writeMessage(epr, (MessageContext)actionContext);
            }

            // if an mtom message module was added to request, this implies that it
            // is required for response using version implied by namespace, i.e.
            // assuming we supported more than one version, which we don't

            if (requestModules.containsKey(WSMTOMMessageModule.class)) {
                isMTOMEnabled = true;
                if (log.isDebugEnabled()) {
                    log.debug("mtom enabled on response action "
                            + responseAction
                            + " defined on epr "
                            + epr.getFragmentIdentifier()
                            + ".");
                }
            } else {
                if (log.isDebugEnabled()) {
                    log.debug("mtom not enabled on response action "
                            + responseAction
                            + " defined on epr "
                            + epr.getFragmentIdentifier()
                            + ".");
                }
            }

            bindingModules = responseEAR.getBindingModules();

        } else {

            // the action is undeclared. we run through default set of binding
            // modules so that required response headers will be generated

            if (actionContext instanceof SOAPFaultContext) {

                // a soap fault was generated before it made it into application
                // layer. no need to bind message just convert to soap fault
                // envelope

                SOAPFault soapFault = ((SOAPFaultContext)actionContext).getSoapFault();
                responseEnvelope = buildSOAPFaultEnvelope(epr, soapFault);

            } else {

                // the only undeclared action we support is the system fault
                // action. throw it back to engine if that's not the case

                if (!actionContext.getAction().equals(SystemFault.ACTION)) {
                    throw new SystemFault(app, "Action " + actionContext.getAction() + " is undefined.");
                }

                // wrap system fault in a soap application fault object which
                // will handle manipulation of fault code, detail and action

                SOAPApplicationFault sf = new SOAPApplicationFault(getSOAPVersion(), (FaultContext)actionContext);
                responseEnvelope = buildSOAPFaultEnvelope(epr, sf);

            }

            isFaultResponse = true;
            bindingModules = epr.getDefaultBindingModules();

        }

        // execute natively implemented binding modules. the version of spec
        // employed is version selected by client, i.e. implied by namespace
        // of request modules

        WSAHandler wsah = WSAHandler.getInstance(wsammIn.getNamespace());
        MessageModules responseModules = actionContext.getMessageModules();
        wsammOut = wsah.handleMyServiceResponse(requestModules, responseModules, responseEnvelope);
        responseModules.putModule(wsammOut);

        // execute binding modules. note that modules natively implemented by
        // this binding are a no-op

        for (BindingModule bm : bindingModules) {
            MessageModule mm = bm.handleMyServiceResponse(requestModules, (Content)responseEnvelope);
            if (mm != null) {
                responseModules.putModule(mm);
            }
        }

        // write all application data passed to us by application as soap
        // header blocks

        if (responseEAR != null) {
            writeApplicationData(responseEAR, responseModules, responseEnvelope);
        }

        // persist the response modules to dbms
        for (MessageModule mm : responseModules.values()) {
            messageModuleDAO.insertMessageModule(wsammOut.getMessageID(), true, mm);
        }

        OMOutputFormat omFormat = new OMOutputFormat();
        omFormat.setSOAP11(getSOAPVersion() == SOAPVersion.SOAP11);
        omFormat.setDoOptimize(isMTOMEnabled);
        omFormat.setCharSetEncoding(Constants.UTF_8);

        String contentType = getContentType(omFormat, null);

        // if replyTo is anonymous uri, then we are returning a response
        // generated by this binding, i.e. a soap fault was thrown before
        // the request was written to receive queue. no jms requestor is
        // blocking. serialize the response to http outputstream

        String replyTo = wsammIn.getReplyTo().getAddress();

        if (replyTo.equals(WSA10.ANONYMOUS_URI)) {

            log.debug("returning soap fault generated by binding");

            HTTPServerResponse response = ((SOAPFaultContext)actionContext).getResponse();
            response.setContentType(contentType);

            // when returning a soap fault, the status code must be set
            // to 500 as required by basic profile

            response.setStatus(Constants.SC_INTERNAL_SERVER_ERROR);

            XMLStreamWriter writer = null;
            try {
                OutputStream out = response.getOutputStream();
                writer = STAXUtils.createXMLStreamWriter(out, Constants.UTF_8);
                ((ElementImpl)responseEnvelope).serializeAndConsume(writer, omFormat);
            } catch (Throwable th) {
                log.error("Error serializing synchronous binding fault response. " + th);
            } finally {
                if (writer != null) {
                    writer.close();
                }
            }

            log.trace("sendMyServiceResponse end");
            return;

        }

        // if replyTo starts with jms, then a jms requestor is blocking for a
        // response. this implies that either engine is returning a response
        // or we failed to bind the request prior to invoking engine, i.e.
        // the request was not consistent with wsdl description

        if (replyTo.startsWith("jms")) {

            // retrieve a refererence to temp queue by name. note that we use
            // this approach rather than dynamically creating using session
            // on sender. activemq overwrites the existing queue with a new
            // destination and registers with mbean, etc ... not what we want

            String tempQueueName = replyTo.substring(replyTo.indexOf(":") + 1);
            Queue temporaryQueue = jmsRequestorFactory.getTemporaryQueue(tempQueueName);

            // if the requestor timed out waiting for response, and cache was
            // soft, then requestor may have been garbage collected, in which
            // case, the temp queue is null

            if (temporaryQueue == null) {
                log.debug("temporary queue was destroyed. discarding anon response.");
                return;
            }

            // stream the response to a temporary file which gets deleted
            // automatically when jvm dies, i.e. no sense in preserving it
            // beyond the life of jvm because servlet response object will
            // no longer exist nor will the blocking jms requestor

            File responseContent = null;
            OutputStream out = null;
            XMLStreamWriter writer = null;
            try {
                responseContent = File.createTempFile("eoa-", ".xml", tempDirectory);
                responseContent.deleteOnExit();
                out = new FileOutputStream(responseContent);
                writer = STAXUtils.createXMLStreamWriter(out, Constants.UTF_8);
                ((ElementImpl)responseEnvelope).serializeAndConsume(writer, omFormat);
            } finally {
                if (out != null) {
                    out.flush();
                    IOUtils.closeQuietly(out);
                }
                if (writer != null) {
                    writer.close();
                }
            }

            log.debug("writing my service response to temp queue");

            JMSSender jmsSender = null;
            try {
                jmsSender = (JMSSender)jmsSenders.borrowObject();
                javax.jms.Message msg = jmsSender.createMessage();
                msg.setStringProperty(Constants.FILE_NAME_PROPERTY, responseContent.getAbsolutePath());
                msg.setBooleanProperty(Constants.GZIP_RESPONSE_PROPERTY, gzipResponse);
                msg.setStringProperty(Constants.CONTENT_TYPE_PROPERTY, contentType);
                msg.setStringProperty(Constants.ACTION_PROPERTY, wsammOut.getAction());
                msg.setBooleanProperty(Constants.IS_FAULT_RESPONSE_PROPERTY, isFaultResponse);
                msg.setJMSCorrelationID(wsammIn.getMessageID());
                jmsSender.send(msg, temporaryQueue, 0, DeliveryMode.NON_PERSISTENT);
            } finally {
                jmsSenders.returnObject(jmsSender);
            }

            log.trace("sendMyServiceResponse end");
            return;

        }

        // if we make it this far, we're returning a response asynchronously

        log.debug("returning my service response asynchronously");

        Map<String, String> headers = new HashMap<String, String>();
        headers.put(Constants.HEADER_CONTENT_TYPE, contentType);

        if (gzipResponse) {
            headers.put(Constants.HEADER_CONTENT_ENCODING, Constants.COMPRESSION_GZIP);
        }

        // typically soap action header is only required for initial message in
        // exchange, which is consistent with soap spec. but because soap spec
        // doesn't address async messaging, we must defer to ws-addressing spec
        // which says header is required for soap 1.1 (it doesn't differentiate
        // between the initial and final message in exchange). note that without
        // soap action header, axis (as of version 1.3), interprets the text/xml
        // media type (i.e. soap 1.1) as REST. which is NOT interoperable

        setSOAPAction(headers);

        // adjust replyTo according to whether or not this is a fault,
        // and whether or not a faultTo address exists

        if (isFaultResponse && wsammIn.getFaultTo() != null) {
            replyTo = wsammIn.getFaultTo().getAddress();
        }

        // add user defined header values which are optionaly defined within
        // binding configuration

        Map<String, String> userDefinedHeaders = httpHeaders.get(wsammOut.getAction());
        if (userDefinedHeaders != null) {
            for (Map.Entry<String, String> entry : userDefinedHeaders.entrySet()) {
                String name = entry.getKey();
                String value = entry.getValue();
                if (headers.get(name) == null) {
                    headers.put(name, value);
                } else {
                    log.debug("not overwriting existing http header '"
                            + name
                            + "' with user defined value '"
                            + value
                            + "'.");
                }
            }
        }

        ClassLoader cloader = factoryContext.getClassLoader();
        RequestEntity re = new HTTPRequestEntity(cloader, responseEnvelope, omFormat, gzipResponse);
        HTTPClientRequest request = new HTTPClientRequest(replyTo, Method.POST, headers, re);

        // attempt to send the response. if module is unable to send, e.g. due
        // to a communication error, etc ..., it throws a transport fault, which
        // propogates back to engine (after being converted to a system fault by
        // spi_sendAction()), i.e. as an indication that receiver did not receive
        // the response

        HTTPTransportModule transportModule = null;
        if (responseEAR == null) {
            transportModule = (HTTPTransportModule)epr.getDefaultTransportModule();
        } else {
            transportModule = (HTTPTransportModule)responseEAR.getTransportModule();
        }

        // note that we set handleRedirects to true which implies that do not
        // have to worry about handling 3xx codes, i.e. if redirects are not
        // successful, module will throw a transport fault

        HTTPClientResponse clientResponse = null;
        try {
            clientResponse = transportModule.doSend(request, true);
        } finally {
            if (clientResponse != null) {
                clientResponse.releaseConnection();
            }
        }

        // if 4xx status is returned, according to eoa spec, we interpret this
        // as response was not 'accepted for processing'. throw system fault
        // back to engine and let it handle it

        int sc = clientResponse.getStatusCode();

        if (sc >= 400 && sc < 500) {

            log.debug("my service response not accepted. throwing system fault");

            throw new SystemFault(app, "Response was not accepted for processing. "
                    + "Receiver returned HTTP response status code "
                    + clientResponse.getStatusCode()
                    + " and reason phrase '"
                    + clientResponse.getStatusReasonPhrase()
                    + "'.");

        }

        // according to eoa spec, any other status is considered accepted
        // even a 5xx status, which implies a soap fault exists in response
        // which eoa spec indicates must be discarded

        if (sc >= 500 && log.isDebugEnabled()) {
            log.debug("5xx status returned in response to my service response. "
                    + " discarding response content, if any.");
        }

        log.trace("sendMyServiceResponse end");

    }

    /*
     * action represents a request addressed to a partner service. message direction is in and
     * all supported meps are valid
     */
    private void sendPartnerServiceRequest(EndpointReference epr, MessageContext messageContext) throws SystemFault {

        log.trace("sendPartnerServiceRequest begin");

        // as required by spec, if an error occurs before we transition to
        // the 'sending+receiving' state, we throw a fault to spi_sendAction
        // method where it will be converted into a system fault and thrown
        // back to engine as an indication action was not accepted by partner

        log.debug("sending partner service request");

        ApplicationRT app = ((EngineRT)epr.getRootComponent()).getApplication();
        MessageModules requestModules = messageContext.getMessageModules();
        WSAMessageModule wsammIn = requestModules.getModule(WSAMessageModule.class);

        EndpointActionReference requestAction = epr.getEndpointActionReference(wsammIn.getAction());

        // attempt to persist the message. if id is not unique, a sql
        // exception is thrown

        try {
            messageDAO.insertMessage(wsammIn.getMessageID(), false);
        } catch (SQLIntegrityConstraintViolationException sve) {
            throw new SystemFault(app, "Request messageID " + wsammIn.getMessageID() + " not unique.");
        } catch (SQLException se) {
            throw new SystemFault(app, "Error persisting request. " + se);
        }

        // convert abstract message instance along with any application data
        // into a version specific soap envelope which represents concrete
        // message instance

        SOAPEnvelope requestEnvelope = writeMessage(epr, (MessageContext)messageContext);

        // currently we support only one version of wsa, but spec allows
        // support for future versions of wsa spec

        WSABindingModule wsabm = requestAction.getBindingModule(WSABindingModule.class);
        WSA wsa = (WSA)wsabm.getSupportedVersions().values().iterator().next();
        WSSpecification.Policy wsaPolicy = wsa.getPolicy().iterator().next();
        Boolean isAnonymousResponses = ((WSA10.Policy)wsaPolicy).isAnonymousResponses();

        EndpointOperationReference eor = (EndpointOperationReference)requestAction.getParent();
        MEP mep = eor.getEndpointOperation().getInterfaceOperation().getMessageExchangePattern();

        // set response epr(s). if policy is non specific regarding anonymous
        // responses, check for a replyTo address set via private policy

        org.bluestemsoftware.specification.wsa.EndpointReference replyTo = null;
        org.bluestemsoftware.specification.wsa.EndpointReference faultTo = null;

        if (mep == MEP.IN_ONLY) {
            replyTo = WSA10.NONE_REPLY_TO;
        } else if (mep == MEP.ROBUST_IN_ONLY) {
            replyTo = WSA10.NONE_REPLY_TO;
            if (isAnonymousResponses == null) {
                String replyToAddress = ((WSA10.Policy)wsaPolicy).getReplyToAddress();
                if (replyToAddress == null) {
                    // otherwise we could block for a fault that may never be sent
                    throw new SystemFault(app, "Anonymous response not supported with 'robust-in-only' MEP");
                } else {
                    faultTo = new org.bluestemsoftware.specification.wsa.EndpointReference(replyToAddress);
                }
                isAnonymousResponses = Boolean.FALSE;
            } else if (isAnonymousResponses == false) {
                String replyToAddress = ((WSA10.Policy)wsaPolicy).getReplyToAddress();
                faultTo = new org.bluestemsoftware.specification.wsa.EndpointReference(replyToAddress);
            } else {
                // otherwise we could block for a fault that may never be sent
                throw new SystemFault(app, "Anonymous response not supported with 'robust-in-only' MEP");
            }
        } else if (mep == MEP.IN_OUT) {
            if (isAnonymousResponses == null) {
                String replyToAddress = ((WSA10.Policy)wsaPolicy).getReplyToAddress();
                if (replyToAddress == null) {
                    replyTo = WSA10.DEFAULT_REPLY_TO;
                    isAnonymousResponses = Boolean.TRUE;
                } else {
                    replyTo = new org.bluestemsoftware.specification.wsa.EndpointReference(replyToAddress);
                    isAnonymousResponses = Boolean.FALSE;
                }
            } else if (isAnonymousResponses == false) {
                String replyToAddress = ((WSA10.Policy)wsaPolicy).getReplyToAddress();
                replyTo = new org.bluestemsoftware.specification.wsa.EndpointReference(replyToAddress);
            } else {
                replyTo = new org.bluestemsoftware.specification.wsa.EndpointReference(WSA10.ANONYMOUS_URI);
            }
        }

        // prime ws-addressing message module with properties which will be
        // used by handler to write headers to envelope

        String to = epr.getEndpoint().getAddress();
        wsammIn.setTo(to);
        wsammIn.setReplyTo(replyTo);
        wsammIn.setFaultTo(faultTo);
        WSAHandler wsah = WSAHandler.getInstance(WSA10.NAMESPACE_URI);
        wsah.handlePartnerServiceRequest(requestModules, requestEnvelope);

        // currently we support only one version of ws-mtom, but spec allows
        // for support of future versions of wsa spec.

        boolean isMTOMEnabled = false;
        WSMTOMBindingModule mtombm = requestAction.getBindingModule(WSMTOMBindingModule.class);
        if (mtombm != null) {
            WSMTOM mtom = (WSMTOM)mtombm.getSupportedVersions().values().iterator().next();
            WSSpecification.Policy mtomPolicy = mtom.getPolicy().iterator().next();
            isMTOMEnabled = ((WSMTOM10.Policy)mtomPolicy).isRequired();
        }

        // if enabled, add message module to set of request modules, which won't
        // actually be used for anything other than auditing, i.e. mtom is auto-
        // detected on partner response using content-type string

        if (isMTOMEnabled) {
            WSMTOMMessageModule mtommmIn = new WSMTOMMessageModule(WSMTOM.WSMTOM10.NAMESPACE_URI);
            requestModules.putModule(mtommmIn);
        }

        // invoke binding modules. note that modules natively implemented by
        // this binding are a no-op. if an error is encountered, module
        // throws a soap fault, which will be propogated to engine

        for (BindingModule bm : requestAction.getBindingModules()) {
            MessageModule mmIn = null;
            try {
                mmIn = bm.handlePartnerServiceRequest(requestModules, (Content)requestEnvelope);
            } catch (BindingFaultRT bf) {
                throw new SystemFault(app, bf.getFaultName(), bf.getFaultReason());
            }
            if (mmIn != null) {
                requestModules.putModule(mmIn);
            }
        }

        // write all application data passed to us by application as soap
        // header blocks

        writeApplicationData(requestAction, requestModules, requestEnvelope);

        // persist the request modules to the dbms

        for (MessageModule mm : requestModules.values()) {
            try {
                messageModuleDAO.insertMessageModule(wsammIn.getMessageID(), false, mm);
            } catch (Exception ex) {
                throw new SystemFault(app, "Error inserting message modules. " + ex);
            }
        }

        // determine if content is encoded by inspecting algorithms for
        // a supported encoding, i.e. just gzip currently

        HyperTextTransferProtocol http = ((HyperTextTransferProtocol)binding.getUnderlyingProtocol());
        QName operationName = eor.getEndpointOperationName();
        String encodings = http.getMessageReferenceHTTPContentEncoding(operationName, Direction.IN);
        boolean gzipRequest = false;
        if (encodings != null) {
            gzipRequest = encodings.contains("gzip");
        }

        // if exchange indicates a response is warranted, inspect algorithms
        // defined on response action with message direction Out

        boolean gzipResponse = false;
        if (mep != MEP.IN_ONLY) {
            encodings = http.getMessageReferenceHTTPContentEncoding(operationName, Direction.OUT);
            if (encodings != null) {
                gzipResponse = encodings.contains("gzip");
            }
        }

        // configure output format used by axiom to serialize the envelope
        // using a specific version of soap and optional optimization

        OMOutputFormat omFormat = new OMOutputFormat();
        omFormat.setSOAP11(getSOAPVersion() == SOAPVersion.SOAP11);
        omFormat.setDoOptimize(isMTOMEnabled);
        omFormat.setCharSetEncoding(Constants.UTF_8);

        // set the soap action value which will be added as an http header
        // if soap 1.1

        Map<String, String> requestHeaders = new HashMap<String, String>();
        String soapAction = setSOAPAction(requestHeaders);

        // get content encoding which will incorporate the soap action value
        // if it's SOAP 1.2

        String requestContentType = getContentType(omFormat, soapAction);
        requestHeaders.put(Constants.HEADER_CONTENT_TYPE, requestContentType);

        // add http headers which indicate encoding employed on request, if
        // any, and encoding supported on response, if any

        if (gzipRequest) {
            requestHeaders.put(Constants.HEADER_CONTENT_ENCODING, Constants.COMPRESSION_GZIP);
        }

        if (gzipResponse) {
            requestHeaders.put(Constants.HEADER_ACCEPT_ENCODING, Constants.COMPRESSION_GZIP);
        }

        // add user defined header values which are optionaly defined within
        // binding configuration

        Map<String, String> userDefinedHeaders = httpHeaders.get(wsammIn.getAction());
        if (userDefinedHeaders != null) {
            for (Map.Entry<String, String> entry : userDefinedHeaders.entrySet()) {
                String name = entry.getKey();
                String value = entry.getValue();
                if (requestHeaders.get(name) == null) {
                    requestHeaders.put(name, value);
                } else {
                    log.debug("not overwriting existing http header '"
                            + name
                            + "' with user defined value '"
                            + value
                            + "'.");
                }
            }
        }

        // attempt to send the request. if module is unable to send, e.g. due
        // to a communication error, etc ..., it throws a transport fault, which
        // we propogate back to engine as a system fault. note that timeout is
        // configured via options on client

        ClassLoader cloader = factoryContext.getClassLoader();
        RequestEntity re = new HTTPRequestEntity(cloader, requestEnvelope, omFormat, gzipRequest);
        HTTPClientRequest request = new HTTPClientRequest(to, Method.POST, requestHeaders, re);
        HTTPClientResponse clientResponse = null;
        SOAPEnvelope responseEnvelope = null;

        try {

            // note that we set handleRedirects to true which implies that do not
            // have to worry about handling 3xx codes, i.e. if redirects are not
            // successful, module will throw a transport fault, i.e. which will
            // be converted to a system fault and rethrown to engine within
            // spi_sendAction method

            log.debug("sending request via http transport module");
            HTTPTransportModule tm = (HTTPTransportModule)requestAction.getTransportModule();
            clientResponse = tm.doSend(request, true);

            // according to spec, if response status is 4xx, the request was
            // not accepted by receiver. we must throw a system fault

            int sc = clientResponse.getStatusCode();
            if (log.isDebugEnabled()) {
                log.debug("client returned status " + sc);
            }

            if (sc >= 400 && sc < 500) {
                String faultReason = clientResponse.getStatusReasonPhrase();
                log.debug("request not accepted. throwing system fault");
                throw new SystemFault(app, faultReason);
            }

            // from this point forward, the request is considered to have been
            // 'accepted for processing' by responding soap node, i.e. we now
            // transition from the 'requesting' state to the 'sending+receiving'
            // state. do NOT throw any faults

            try {

                log.debug("request was accepted for processing");

                responseEnvelope = parseHTTPClientResponse(clientResponse);

                if (mep == MEP.IN_ONLY || (responseEnvelope == null && mep == MEP.ROBUST_IN_ONLY)) {
                    log.debug("mep is in-only or mep is robust-in-only or response null");
                    log.trace("sendPartnerServiceRequest end");
                    return;
                }

                // if we expected an anonymous response, but the response was null, all
                // we can do is log an error and return

                if (isAnonymousResponses && responseEnvelope == null) {
                    log.error("anonymous response expected, when invoking partner"
                            + " action addressed to "
                            + to
                            + ", but response was empty.");
                    log.trace("sendPartnerServiceRequest end");
                    return;
                }

                // if we expect a non-anonymous response, we discard any content
                // which may exist as required by eoa spec including server
                // errors, e.g. 500+, which should also be returned async (except
                // soap protocol faults which are always returned anon.) note
                // that axis returns a 200 status with content type text/xml in
                // response to an async request (as of version 1.3). attempting
                // to process the response content could be problematic across
                // vendors, which is why the eoa spec says to ignore it

                boolean isSOAPProtocolFault = isSOAPProtocolFault(responseEnvelope);

                if (!isAnonymousResponses && !isSOAPProtocolFault) {
                    log.debug("partner response will be returned aysnchronously");
                    log.trace("sendPartnerServiceRequest end");
                    return;
                }

            } catch (Throwable th) {

                // we were expecting a response, but we failed to parse it.
                // all we can do here is log an error and move on

                log.error("error parsing anonymous partner response sent in response to action "
                        + wsammIn.getAction()
                        + " addressed to endpoint "
                        + wsammIn.getTo()
                        + ". "
                        + th);

                return;

            }

            log.debug("partner response returned anonymously");

            // pass response to receivePartnerServiceResponse method where
            // response will be processed and passed to engine on a separate
            // thread

            receivePartnerServiceResponse(null, null, clientResponse, epr, null, responseEnvelope);

        } finally {

            // response envelope has been serialized to file system by this
            // point, so it's safe to close the underlying stream, i.e. by
            // releasing connection

            if (clientResponse != null) {
                clientResponse.releaseConnection();
            }

        }

        log.trace("sendPartnerServiceRequest end");
    }

    private SOAPEnvelope parseHTTPClientResponse(HTTPClientResponse clientResponse) throws IOException {

        // as of version 1.3, in some situations, axis writes garbage to
        // http response body and returns without a content-type header
        // when response status is accepted (202). so if status is accepted,
        // do not attempt to parse, i.e. because decendant will throw an
        // unsupported media type exception

        InputSource responseContent = null;
        if (clientResponse.getStatusCode() == 202) {
            return null;
        } else {
            responseContent = clientResponse.getContent();
        }

        if (responseContent == null) {
            return null;
        }

        Map<String, String> responseHeaders = clientResponse.getHeaders();

        // retrieve input stream from request. check if content is encoded
        // using gzip, which is the only encoding we support currently

        InputStream is = null;
        String ce = responseHeaders.get(Constants.HEADER_CONTENT_ENCODING);
        if (ce != null) {
            if (ce.equalsIgnoreCase("gzip")) {
                try {
                    is = new GZIPInputStream(responseContent.getByteStream());
                } catch (IOException ie) {
                    if (managedConfiguration.isEmbeddStackTraceInFaults()) {
                        throw new SOAPReceiverFault(getSOAPVersion(), "Error decoding content.", ie);
                    } else {
                        log.warn("Error decoding content.", ie);
                        throw new SOAPReceiverFault(getSOAPVersion(), "Error decoding content. " + ie.getMessage());
                    }
                }
            } else {
                throw new SOAPSenderFault(getSOAPVersion(), "Content-Encoding '" + ce + "' not supported.");
            }
        } else {
            is = responseContent.getByteStream();
        }

        // extract media type from content type header. if media type is un-
        // expected, descendant throws a soap fault

        String responseCType = responseHeaders.get(Constants.HEADER_CONTENT_TYPE);
        responseCType = responseCType == null ? Constants.DEFAULT_CONTENT_TYPE : responseCType;
        String responseMediaType = getMediaType(responseCType);

        // if media type implies mtom, then we create an attachments object, which
        // leverages the axiom mtom implementation

        Attachments attachments = null;
        if (responseMediaType.equals(Constants.SOAP_MTOM_MEDIA_TYPE)) {
            boolean cacheAttachments = managedConfiguration.getLargeAttachmentSize() > -1;
            String threshold = Integer.toString(managedConfiguration.getLargeAttachmentSize());
            attachments = new Attachments(is, responseCType, cacheAttachments, TEMP_DIR, threshold);
        }

        return buildSOAPEnvelope(is, attachments, responseContent.getEncoding());

    }

    /*
     * parses optional char set parameter from content type, e.g.
     * application/soap+xml;action="http://foo"[;charset=utf-8]
     */
    private String getCharacterSetEncoding(HTTPServerRequest request) {

        String answer = request.getContentType();

        if (answer == null) {
            return Constants.UTF_8;
        }

        // locate name value pair. if undefined, default value to utf-8
        int index = answer.indexOf("charset");
        if (index == -1) {
            return Constants.UTF_8;
        }

        // name value pair may be the last pair or an intermediate pair
        index = answer.indexOf("=", index);
        if (answer.indexOf(";", index) == -1) {
            answer = answer.substring(index + 1, answer.length()).trim();
        } else {
            answer = answer.substring(index + 1, answer.indexOf(";", index)).trim();
        }

        // trim double quotes which surround encoding, if any
        if (answer.startsWith("\"") && answer.endsWith("\"")) {
            answer = answer.substring(1, answer.length());
        }

        return answer.trim();

    }

    public class WorkImpl implements Work {

        private SOAPFault sf;
        private EndpointReference epr;
        private EndpointActionReference ear;
        private MessageModules requestModules;

        public WorkImpl(SOAPFault sf, EndpointReference epr, EndpointActionReference ear,
                MessageModules requestModules) {
            this.sf = sf;
            this.epr = epr;
            this.ear = ear;
            this.requestModules = requestModules;
        }

        public void doWork() {

            // if ws-addressing was not engaged, this is considered fatal,
            // i.e. request modules did not execute. note that basic profile
            // (R1151) prohibits us from returning fault response over the
            // back channel if reply is anonymous

            if (sf instanceof WSAddressingFault) {

                SOAPEnvelope faultEnvelope = buildSOAPFaultEnvelope(epr, sf);
                OMOutputFormat omFormat = new OMOutputFormat();
                omFormat.setDoOptimize(false);
                omFormat.setCharSetEncoding(Constants.UTF_8);
                omFormat.setAutoCloseWriter(true);
                omFormat.setSOAP11(sf.getSoapVersion() == SOAPVersion.SOAP11);
                String contentType = getContentType(omFormat, null);

                WSAHandler wsaHandler = WSAHandler.getInstance(WSA10.NAMESPACE_URI);
                WSAMessageModule wsammIn = ((WSAddressingFaultImpl)sf).getOffendingMessageModule();
                String mid = wsammIn.getMessageID();

                String replyTo = null;
                if (wsammIn.getFaultTo() == null) {
                    replyTo = wsammIn.getReplyTo().getAddress();
                } else {
                    replyTo = wsammIn.getReplyTo().getAddress();
                }

                wsaHandler.handleWSAFaultResponse(replyTo, mid, sf.getAction(), faultEnvelope);

                Map<String, String> headers = new HashMap<String, String>();
                headers.put(Constants.HEADER_CONTENT_TYPE, contentType);
                ClassLoader cloader = factoryContext.getClassLoader();
                RequestEntity re = new HTTPRequestEntity(cloader, faultEnvelope, omFormat, false);
                HTTPClientRequest hcr = new HTTPClientRequest(replyTo, Method.POST, headers, re);

                HTTPTransportModule tm = null;
                if (ear == null) {
                    tm = (HTTPTransportModule)epr.getDefaultTransportModule();
                } else {
                    tm = (HTTPTransportModule)ear.getTransportModule();
                }

                // return the response asynchronously. if there's an error, there's not
                // much we can do, other than log it, i.e. the http response has already
                // been accepted

                HTTPClientResponse clientResponse = null;
                try {
                    clientResponse = tm.doSend(hcr, true);
                } catch (Throwable tf) {
                    if (tf instanceof TransportFault) {
                        log.error("caught throwable returning async wsa fault '"
                                + sf.getFaultReason()
                                + "' in response to message "
                                + wsammIn.getMessageID()
                                + ". "
                                + tf.getMessage());
                    } else {
                        log.error("caught throwable returning async wsa fault '"
                                + sf.getFaultReason()
                                + "' in response to message "
                                + wsammIn.getMessageID()
                                + ". "
                                + tf);
                    }
                } finally {
                    if (clientResponse != null) {
                        clientResponse.releaseConnection();
                    }
                }

                return;

            }

            // ws-addressing was engaged, we can run the binding fault response through
            // my service response logic, i.e. so that response modules will be processed

            try {
                WSAMessageModule wsammOut = new WSAMessageModule(sf.getAction());
                WSAMessageModule wsammIn = requestModules.getModule(WSAMessageModule.class);
                wsammOut.setRelatesTo(wsammIn.getMessageID());
                MessageModules responseModules = new MessageModules(wsammOut);
                SOAPFaultContext sfc = new SOAPFaultContext(null, requestModules, responseModules, sf);
                sendMyServiceResponse(epr, sfc);
            } catch (Throwable thr) {
                log.error("caught throwable returning async fault response. " + thr);
            }

        }

    }

    private void checkRequiredHTTPHeaders(String action, HTTPClientResponse response, HTTPServerRequest request) throws SOAPFault {

        if (binding.isReusable()) {
            return; // no abstract header decls allowed
        }

        BindingAction ba = binding.getBindingAction(action);
        if (ba == null) {
            return;
        }

        HyperTextTransferProtocol http = (HyperTextTransferProtocol)binding.getUnderlyingProtocol();

        if (ba instanceof SOAPBindingMessageReference) {
            QName operationName = ((SOAPBindingMessageReference)ba).getParent().getName();
            Direction direction = ((SOAPBindingMessageReference)ba).getDirection();
            for (HTTPHeader httpHeader : http.getMessageReferenceHTTPHeaders(operationName, direction)) {
                String name = httpHeader.getName();
                String value = null;
                if (request == null) {
                    value = response.getHeaders().get(name);
                } else {
                    value = request.getHeader(name);
                }
                if (value == null) {
                    if (httpHeader.isRequired()) {
                        throw new SOAPSenderFault(getSOAPVersion(), "Required http header '"
                                + name
                                + "' is missing.");
                    }
                } else {
                    String expected = configuration.getMessageReferenceHTTPHeaderValue(operationName, direction,
                            name);
                    if (expected != null) {
                        if (!expected.equals(value)) {
                            throw new SOAPSenderFault(getSOAPVersion(), "HTTP header '"
                                    + name
                                    + "' failed to match expected value '"
                                    + expected
                                    + "'.");
                        }
                    }
                }
            }
        } else {
            BindingFault bf = ((BindingFaultReference)ba).getFault();
            QName faultName = bf.getFaultName();
            for (HTTPHeader httpHeader : http.getFaultHTTPHeaders(faultName)) {
                String name = httpHeader.getName();
                String value = null;
                if (request == null) {
                    value = response.getHeaders().get(name);
                } else {
                    value = request.getHeader(name);
                }
                if (value == null) {
                    if (httpHeader.isRequired()) {
                        throw new SOAPSenderFault(getSOAPVersion(), "Required http header '"
                                + name
                                + "' is missing.");
                    }
                } else {
                    String expected = configuration.getFaultHTTPHeaderValue(faultName, name);
                    if (expected != null) {
                        if (!expected.equals(value)) {
                            throw new SOAPSenderFault(getSOAPVersion(), "HTTP header '"
                                    + name
                                    + "' failed to match expected value '"
                                    + expected
                                    + "'.");
                        }
                    }
                }
            }
        }

    }

    private void cacheRequiredHTTPHeaders(String action, BindingAction ba) throws SOAPBindingException {

        Map<String, String> headerMap = new HashMap<String, String>();
        httpHeaders.put(action, headerMap);

        HyperTextTransferProtocol http = (HyperTextTransferProtocol)binding.getUnderlyingProtocol();

        if (ba instanceof SOAPBindingMessageReference) {
            QName operationName = ((SOAPBindingMessageReference)ba).getParent().getName();
            Direction direction = ((SOAPBindingMessageReference)ba).getDirection();
            for (HTTPHeader httpHeader : http.getMessageReferenceHTTPHeaders(operationName, direction)) {
                String name = httpHeader.getName();
                String value = configuration.getMessageReferenceHTTPHeaderValue(operationName, direction, name);
                if (value == null) {
                    if (httpHeader.isRequired()) {
                        throw new SOAPBindingException("Value for required http header '"
                                + name
                                + "' not defined within binding provider configuration.");
                    }
                } else {
                    headerMap.put(name, value);
                }
            }
        } else {
            BindingFault bf = ((BindingFaultReference)ba).getFault();
            QName faultName = bf.getFaultName();
            for (HTTPHeader httpHeader : http.getFaultHTTPHeaders(faultName)) {
                String name = httpHeader.getName();
                String value = configuration.getFaultHTTPHeaderValue(faultName, name);
                if (value == null) {
                    if (httpHeader.isRequired()) {
                        throw new SOAPBindingException("Value for required http header '"
                                + name
                                + "' not defined within binding provider configuration.");
                    }
                } else {
                    headerMap.put(name, value);
                }
            }
        }

    }

}
